import asyncio

import pytest

from chancy import Chancy, Worker, Queue, job


@job()
async def quick_job():
    await asyncio.sleep(1)


@pytest.mark.asyncio
async def test_regression_48(chancy: Chancy, worker: Worker):
    """
    When queues are configured with `concurrency=1`, it will always wait for
    `polling_interval` before fetching additional jobs, even if it's likely
    that there are more jobs to process.
    """
    await chancy.declare(
        Queue(
            "normal",
            concurrency=1,
            polling_interval=5,
            executor=Chancy.Executor.Async,
        )
    )

    all_refs = [
        ref
        async for refs in chancy.push_many(
            [quick_job.job.with_queue("normal")] * 5
        )
        for ref in refs
    ]

    # This will take 25 seconds, as the worker will process one job, then
    # wait 5 seconds before fetching the next job.
    with pytest.raises(asyncio.TimeoutError):
        await chancy.wait_for_jobs(all_refs, timeout=10)


@pytest.mark.asyncio
async def test_regression_48_eager(chancy: Chancy, worker: Worker):
    """
    When queues are configured with `concurrency=1`, it will always wait for
    `polling_interval` before fetching additional jobs, even if it's likely
    that there are more jobs to process.
    """

    await chancy.declare(
        Queue(
            "eager",
            concurrency=1,
            polling_interval=5,
            executor=Chancy.Executor.Async,
            eager_polling=True,
        )
    )

    all_refs = [
        ref
        async for refs in chancy.push_many(
            [quick_job.job.with_queue("eager")] * 5
        )
        for ref in refs
    ]

    # This should take roughly 5 seconds, as the worker will process one job,
    # then immediately fetch the next job.
    await chancy.wait_for_jobs(all_refs, timeout=10)



import asyncio
from typing import Any

import pytest

from chancy import Chancy, Worker, Queue, QueuedJob, job
from chancy.plugin import Plugin


class RunCounter(Plugin):
    @staticmethod
    def get_identifier() -> str:
        return "run_counter"

    def __init__(self):
        super().__init__()
        self.runs = 0

    def on_job_completed(
        self,
        *,
        worker: "Worker",
        job: QueuedJob,
        exc: Exception | None = None,
        result: Any | None = None,
    ) -> QueuedJob:
        self.runs += 1
        return job


@job()
async def long_running_job():
    await asyncio.sleep(5)


@pytest.mark.asyncio
@pytest.mark.parametrize("chancy", [{"plugins": [RunCounter()]}], indirect=True)
async def test_regression_51(chancy: Chancy, worker: Worker):
    """
    Test that pushing a job with a unique key while the same job is running
    does not result in concurrent execution of the job.

    Uses a shim plugin to count the # of jobs actually run, since the issue
    causes ID re-use, we can't rely on job IDs to count runs.
    """
    await chancy.declare(Queue("default"))
    # If working properly, ref_one and ref_two should be the same. However, if
    # we wait until ref_one is actively running, then ref_two will be allowed
    # to be pushed again, and we'll get a new reference ID.
    await chancy.push(
        long_running_job.job.with_unique_key("unique_long_running_job")
    )
    await asyncio.sleep(1)
    await chancy.push(
        long_running_job.job.with_unique_key("unique_long_running_job")
    )
    await asyncio.sleep(10)

    assert chancy.plugins[RunCounter.get_identifier()].runs == 1



from psycopg import sql

from chancy.migrate import Migration


class QueueEagerPolling(Migration):
    """
    Adds the eager polling field to queues.
    """

    async def up(self, migrator, cursor):
        await cursor.execute(
            sql.SQL(
                """
                ALTER TABLE {queues} 
                ADD COLUMN IF NOT EXISTS eager_polling BOOLEAN NOT NULL DEFAULT FALSE
                """
            ).format(queues=sql.Identifier(f"{migrator.prefix}queues"))
        )

    async def down(self, migrator, cursor):
        await cursor.execute(
            sql.SQL(
                """
                ALTER TABLE {queues} 
                DROP COLUMN IF EXISTS eager_polling
                """
            ).format(queues=sql.Identifier(f"{migrator.prefix}queues"))
        )
